/*
$Log: menu.h,v $
Revision 3.2.1.1  2000/12/18 08:21:48  shelton
Import slightly modified sources of ITK yui project. made by Sergey Rozenfeld.

Revision 3.1.15.1  2000/12/18 08:21:48  shelton
Import original yui sources, patched only for succesful compilation on
FreeBSD

Revision 1.5  1999/10/23 20:26:46  serg
terminfo support

Revision 1.4  1998/05/04 17:07:46  serg
       X-window system
CVS:
CVS:

Revision 1.3  1998/04/29 00:33:02  serg
    -Wall  , 
  ,     
  shadowMap
CVS:
CVS:

Revision 1.2  1998/04/15 17:40:37  serg
Ported to Solaris-2.6-x86
CVS:
CVS:

Revision 1.1  1997/08/01 17:45:25  paul
initial

*/
#ifndef _MENU_H
#define _MENU_H

#include "hashcode.h"
#include "listbox.h"

#define COMMAND		0
#define MENU		1
#define SEPARATOR	2

class Menu;

struct menuItem
{
   u_char type;
   Lang *label;
   Lang *status;
   char *key;
   char *help;
   union {
       long command;
       Menu *menu;
   };
   menuItem( Lang *msg, Lang *s=0, char *h=0, long command=HASH_cmOK );
   menuItem( Lang *msg, Lang *s=0, char *h=0, Menu *m=0 );
   menuItem( char *l=0, char *s=0, char *h=0, long command=HASH_cmOK );
   menuItem( char *l=0, char *s=0, char *h=0, Menu *m=0 );
   ~menuItem();
   void init( char *h );
   u_char hotkey();
};

extern long menuKeyMap[];
extern Keymap MenuKeyMap;

/*---------------------------------------------------------------*/
class menuColl : public Collection
{
protected:
   void freeItem( void *item ) { delete (menuItem*)item; }
public:
   menuColl() : Collection(10,10) {;}
   ~menuColl() { freeAll(); }
};

class Menu : virtual public Window
{
friend class MenuManager;

protected:
   Rect orig;
   Point langPoint[LANGS];
   menuColl *mcoll;
   int vert, offsetRect, width;
//   menuColor *color;
   void beforeExec();
   void afterExec( long key );
   void startInit( int Box, Lang *tit );
   void setRect();

public:
   Menu( menuColl *it, Point coor, Lang *tit=0, int Vert=1, int Auto=1, int Box=1 );
   Menu( menuColl *it, Point coor1, Point coor2, Lang *tit=0, int Vert=1, int Auto=1, int Box=1 );
   ~Menu();
   virtual void reallocScrMap();
   virtual int isType(long typ);
   virtual void setColors( int type );
   int autoOpen;
   static int state;
   static int curstate;
   int oldState;
   int init( void *data=0 );
   long handleKey( long key, void *&ptr );
   void changeRect( int offset );
   long changeCurrent( long key, int flag=0 );
   int draw( int Redraw=0, int sy=0, int sx=0 );
   void drawStatus();
   void moveCursor();
   void fill( KeyHolder *kmap=0 );
   void setHotKeys( char *mas );
   Point getPoint();
   void *varptr;
};

class MenuManager : virtual public Window
{
protected:
   int offset;
//   menuColor *color;
   virtual void beforeExec();
   virtual void afterExec( long key );

public:
   MenuManager();
   virtual void reallocScrMap();
   virtual int 	isType(long typ);
   virtual void setColors( int type );
   int init( void *data=0 );
   void select( Menu *m, int index=-1 );
   void remove( Menu *m );
   virtual long handleKey( long key, void *&ptr );
   virtual int draw( int Redraw=0, int sy=0, int sx=0 );
   virtual void setScrMap();
   virtual void moveCursor();
   int sepchar;
   int canSwitch;
   char hotKeys[32];
   int hotKey;
   char Visible, visible;
   int isHotKey( long key );
};

class SharedMenuCollection : public SortedPtrCollection
{
protected:
   void freeItem( void *item ) { delete (Menu*)item; }
public:
   SharedMenuCollection() { shouldDelete=1; }
   ~SharedMenuCollection() { freeAll(); }
};

extern SharedMenuCollection sharedMenu;

#endif
